# jwt_full/joseheader.nim
# Copyright 2021 Jonathan Bernard

## ====================================
##             joseheader
## ====================================
##
## ------------------------------------
##            JOSE Header
## ------------------------------------
##
## Data structures and utility procedures for working with JOSE headers.  This
## module follows the common pattern in this library of providing direct access
## to all of the header parameters via the `[]` syntax, as well as convencience
## accessors for the publicly-defined header parameters.

import std/json, std/options, std/sequtils, std/strutils

import ./encoding, ./jwa, ./jwk
import ./private/jsonutils

type
  JoseHeader* = object
    rawB64: string
    json: JsonNode
    alg: Option[JwtAlgorithm]
    jku: Option[string]
    jwk: Option[JWK]
    kid: Option[string]
    typ: Option[string]
    cty: Option[string]
    crit: Option[string]
    x5u: Option[string]
    x5c: Option[seq[string]]
    x5t: Option[string]
    x5tS256: Option[string]

func rawB64*(h: JoseHeader): string = h.rawB64
func alg*(h: JoseHeader): Option[JwtAlgorithm] = h.alg
func jku*(h: JoseHeader): Option[string] = h.jku
func jwk*(h: JoseHeader): Option[JWK] = h.jwk
func kid*(h: JoseHeader): Option[string] = h.kid
func typ*(h: JoseHeader): Option[string] = h.typ
func cty*(h: JoseHeader): Option[string] = h.crit
func crit*(h: JoseHeader): Option[string] = h.crit
func x5u*(h: JoseHeader): Option[string] = h.x5u
func x5c*(h: JoseHeader): Option[seq[string]] = h.x5c
func x5t*(h: JoseHeader): Option[string] = h.x5t
func x5tS256*(h: JoseHeader): Option[string] = h.x5tS256

func `[]`*(h: JoseHeader, key: string): Option[JsonNode] =
  if h.json.hasKey(key): some(h.json[key])
  else: none[JsonNode]()

proc `$`*(header: JoseHeader): string = header.rawB64

proc initJoseHeader*(n: JsonNode): JoseHeader =
  return JoseHeader(
    rawB64: b64UrlEncode($n),
    json: n,
    alg: if n.hasKey("alg"): some(parseEnum[JwtAlgorithm](n["alg"].getStr))
         else: none[JwtAlgorithm](),
    jku: n.optStrVal("jku"),
    jwk: if n.hasKey("jwk"): some(initJwk(n["jwk"]))
         else: none[JWK](),
    kid: n.optStrVal("kid"),
    typ: n.optStrVal("typ"),
    cty: n.optStrVal("cty"),
    crit: n.optStrVal("crit"),
    x5u: n.optStrVal("x5u"),
    x5c: if n.hasKey("x5c"): some(n["x5c"].getElems.mapIt(it.getStr))
         else: none[seq[string]](),
    x5t: n.optStrVal("x5t"),
    x5tS256: n.optStrVal("x5tS256"))

proc initJoseHeader*(encoded: string): JoseHeader =
  let decoded = b64UrlDecode(encoded)
  result = initJoseHeader(parseJson(decoded))
  result.rawB64 = encoded

proc initJoseHeader*(
    alg: Option[JwtAlgorithm] = some(HS256),
    jku: Option[string] = none[string](),
    jwk: Option[JWK] = none[JWK](),
    kid: Option[string] = none[string](),
    typ: Option[string] = none[string](),
    cty: Option[string] = none[string](),
    crit: Option[string] = none[string](),
    x5u: Option[string] = none[string](),
    x5c: Option[seq[string]] = none[seq[string]](),
    x5t: Option[string] = none[string](),
    x5tS256: Option[string] = none[string](),
    moreParams: seq[tuple[k, v: string]] = @[]
  ): JoseHeader =

  let json = newJObject()
  for p in moreParams: json[p.k] = %p.v

  if alg.isSome: json["alg"] = %alg.get
  if jku.isSome: json["jku"] = %jku.get
  if jwk.isSome: json["jwk"] = %jwk.get
  if kid.isSome: json["kid"] = %kid.get
  if typ.isSome: json["typ"] = %typ.get
  if cty.isSome: json["cty"] = %cty.get
  if crit.isSome: json["crit"] = %crit.get
  if x5u.isSome: json["x5u"] = %x5u.get
  if x5c.isSome: json["x5c"] = %x5c.get
  if x5t.isSome: json["x5t"] = %x5t.get
  if x5tS256.isSome: json["x5tS256"] = %x5tS256.get

  return initJoseHeader(json)

proc combine*(a, b: JoseHeader): JoseHeader =
  let json = newJObject()

  for k, v in a.json: json[k] = v

  for k, v in b.json:
    if json.hasKey(k):
      raise newException(ValueError,
        "Duplicate key '" & k & "' found when combining JoseHeaders. This " &
        "will result in an invalid JWS, JWE, or JWT.")
    json[k] = v

  return initJoseHeader(json)
