# jwt_full/jwk.nim
# Copyright 2021 Jonathan Bernard

## ==============================
##             jwk
## ==============================
##
## ------------------------------
## JSON Web Keys (JWK) - RFC 7518
## ------------------------------
##
## Nim implementation of the data structures and constants defined in RFC 7518.

import std/json, std/options, std/sequtils, std/strutils

import ./jwa
import ./private/jsonutils

type
  JwkKeyType* = enum EcPublic, EcPrivate, RsaPublic, RsaPrivate, Octet

  JwkUse* = enum jwkuSignature = "sig", jwkuEncrypt = "enc"

  JwkOps* = enum jwkopSign = "sign", jwkopVerify = "verify",
                 jwkopEncrypt = "encrypt", jwkopDecrypt = "decrypt",
                 jwkopWrapKey = "wrapKey", jwkopUnwrapKey = "unwrapKey",
                 jwkopDeriveKey = "deriveKey", jwkopDeriveBits = "deriveBits"

  EcPubKey = object of RootObj
    crv*: EcCurve
    x*: string
    y*: Option[string]

  EcPrvKey* = object of EcPubKey
    d*: string

  RsaPubKey* = object of RootObj
    n*: string
    e*: string

  RsaOtherPrime* = object
    r*: string
    d*: string
    t*: string

  RsaPrvKey* = object of RsaPubKey
    d*: string
    p*: Option[string]
    q*: Option[string]
    dp*: Option[string]
    dq*: Option[string]
    qi*: Option[string]
    oth*: Option[seq[RsaOtherPrime]]

  OctetKey* = object
    k*: string

  JWK* = object
    json: JsonNode
    kty: string
    use: Option[string]
    key_ops: Option[string]
    alg: Option[string]
    kid: Option[string]
    x5u: Option[string]
    x5c: Option[string]
    x5t: Option[string]
    x5tS256: Option[string]

    case keyKind: JwkKeyType
    of EcPublic: ecPub: EcPubKey
    of EcPrivate: ecPrv: EcPrvKey
    of RsaPublic: rsaPub: RsaPubKey
    of RsaPrivate: rsaPrv: RsaPrvKey
    of Octet: octKey: OctetKey

  JwkSet* = seq[JWK]

# Read-only public accessors to JWK members
# -----------------------------------------

func kty*(jwk: JWK): string = jwk.kty
func use*(jwk: JWK): Option[string] = jwk.use
func key_ops*(jwk: JWK): Option[string] = jwk.key_ops
func alg*(jwk: JWK): Option[string] = jwk.alg
func kid*(jwk: JWK): Option[string] = jwk.kid
func x5u*(jwk: JWK): Option[string] = jwk.x5u
func x5c*(jwk: JWK): Option[string] = jwk.x5c
func x5t*(jwk: JWK): Option[string] = jwk.x5t
func x5tS256*(jwk: JWK): Option[string] = jwk.x5tS256
func keyKind*(jwk: JWK): JwkKeyType = jwk.keyKind
func ecPub*(jwk: JWK): EcPubKey = jwk.ecPub
func ecPrv*(jwk: JWK): EcPrvKey = jwk.ecPrv
func rsaPub*(jwk: JWK): RsaPubKey = jwk.rsaPub
func rsaPrv*(jwk: JWK): RsaPrvKey = jwk.rsaPrv
func octKey*(jwk: JWK): OctetKey = jwk.octKey

func `[]`*(jwk: JWK, key: string): Option[JsonNode] =
  if jwk.json.hasKey(key): some(jwk.json[key])
  else: none[JsonNode]()

# Public Parsing Functions
# ------------------------

func parseEcPubKey*(n: JsonNode): EcPubKey =
  EcPubKey(
    crv: parseEnum[EcCurve](n.reqStrVal("crv")),
    x: n.reqStrVal("x"),
    y: n.optStrVal("y"))

func parseEcPrvKey*(n: JsonNode): EcPrvKey =
  let pk = parseEcPubKey(n)
  EcPrvKey(crv: pk.crv, x: pk.x, y: pk.y, d: n.reqStrVal("d"))

func parseRsaPubKey*(n: JsonNode): RsaPubKey =
  RsaPubKey(
    n: n.reqStrVal("n"),
    e: n.reqStrVal("e"))

func parseRsaOtherPrime(n: JsonNode): RsaOtherPrime =
  RsaOtherPrime(
    r: n.reqStrVal("r"),
    d: n.reqStrVal("d"),
    t: n.reqStrVal("t"))

func parseRsaPrvKey*(n: JsonNode): RsaPrvKey =
  let pk = parseRsaPubKey(n)
  RsaPrvKey(n: pk.n, e: pk.e,
    d: n.reqStrVal("d"),
    p: n.optStrVal("p"),
    q: n.optStrVal("q"),
    dp: n.optStrVal("dp"),
    dq: n.optStrVal("dq"),
    qi: n.optStrVal("qi"),
    oth:
      if n.hasKey("oth"):
        some(n["oth"].getElems.mapIt(parseRsaOtherPrime(it)))
      else: none[seq[RsaOtherPrime]]()
  )


func parseOctetKey*(n: JsonNode): OctetKey =
  OctetKey(k: n.reqStrVal("k"))

func initJwk*(n: JsonNode): JWK =
  # TODO: documentation, example, handle encrypted keys (JWEs)
  case n["kty"].getStr(""):

  of $ktyEC:
    if n.hasKey("d"):
      result = JWK(
        kty: $ktyEC,
        keyKind: EcPrivate,
        ecPrv: parseEcPrvKey(n))
    else:
      result = JWK(
        kty: $ktyEC,
        keyKind: EcPublic,
        ecPub: parseEcPubKey(n))

  of $ktyRSA:
    if n.hasKey("d"):
      result = JWK(
        kty: $ktyRSA,
        keyKind: RsaPrivate,
        rsaPrv: parseRsaPrvKey(n))
    else:
      result = JWK(
        kty: $ktyRSA,
        keyKind: RsaPublic,
        rsaPub: parseRsaPubKey(n))

  of $ktyOctet:
    result = JWK(
      kty: $ktyOctet,
      keyKind: Octet,
      octKey: parseOctetKey(n))

  else: raise newException(ValueError,
    "Unrecognized or missing kty: '" & n["kty"].getStr("") & "'")

  result.json = n
  result.use = n.optStrVal("use")
  result.key_ops = n.optStrVal("key_ops")
  result.alg = n.optStrVal("alg")
  result.kid = n.optStrVal("kid")
  result.x5u = n.optStrVal("x5u")
  result.x5c = n.optStrVal("x5c")
  result.x5t = n.optStrVal("x5t")
  result.x5tS256 = n.optStrVal("x5t#S256")

func initJwkSet*(n: JsonNode): JwkSet =
  # TODO: documentation, examples, handled encrypted set (JWE)
  if not n.hasKey("keys") or n["keys"].kind != JArray:
    raise newException(ValueError, "JWK Set is missing the 'keys' member, " &
      "or 'keys' is not an array.")

  return n["keys"].getElems.mapIt(initJwk(it))
