import bearssl, bearssl_pkey_decoder

import ../../encoding
import ../../jwa
import ../../jwk

import ./hash
import ./pem

type
  EcPublicKeyObj = object
    curve*: EcCurve
    q*: string
    bearKey: EcPublicKey

  EcPrivateKeyObj = object
    curve*: EcCurve
    x*: string
    bearKey: EcPrivateKey

func toBearSslCurveConst(curve: EcCurve): int32 =
  result = case curve:
    of P256: EC_secp256r1
    of P384: EC_secp384r1
    of P521: EC_secp521r1

func fromBearSslCurveConst(curve: int32): EcCurve =
  result = case curve:
    of EC_secp256r1: P256
    of EC_secp384r1: P384
    of EC_secp521r1: P521
    else: P256

func initEcPublicKeyObj(curve: EcCurve, q: string): EcPublicKeyObj =
  result = EcPublicKeyObj(curve: curve, q: q)
  result.bearKey.curve = curve.toBearSslCurveConst
  result.bearKey.q = cast[ptr byte](result.q.cstring)
  result.bearKey.qlen = uint q.len

func initEcPrivateKeyObj(curve: EcCurve, x: string): EcPrivateKeyObj =
  result = EcPrivateKeyObj(curve: curve, x: x)
  result.bearKey.curve = curve.toBearSslCurveConst
  result.bearKey.x = cast[ptr byte](result.x.cstring)
  result.bearKey.xlen = uint x.len

proc toEcPublicKey(jwk: JWK): EcPublicKeyObj =
  ## Convert an ECDSA public key in JWK format to the wrapper for BearSSL's
  ## ECPublicKey struct.

  # Confusingly, the JWK standard and BearSSL use contradictory names for EC
  # key components. In JWK parlance, x and y (optional) are the public key
  # components. BearSSL only requires one public point, and calls this q.

  let keyObj = if jwk.keyKind == EcPublic: jwk.ecPub
               else: jwk.ecPrv

  return initEcPublicKeyObj(
    curve = keyObj.crv,
    q = b64UrlDecode(keyObj.x))

proc toEcPublicKey(pem: string): EcPublicKeyObj =
  var pkCtx: PkeyDecoderContext
  pkeyDecoderInit(addr pkCtx)
  decodePem(pem, "PUBLIC KEY", addr pkCtx,
    cast[BearPemDecoderCallback](pkeyDecoderPush))

  let k = pkCtx.key.ec
  return initEcPublicKeyObj(k.curve.fromBearSslCurveConst, $cast[cstring](k.q))

proc toEcPrivateKey(jwk: JWK): EcPrivateKeyObj =
  ## Convert an ECDSA private key in JWK format to the wrapper for BearSSL's
  ## ECPrivateKey struct.

  # Confusingly, the JWK standard and BearSSL use contradictory names for EC
  # key components. In JWK parlance, d is the private key components. BearSSL
  # calls this x.
  return initEcPrivateKeyObj(
    curve = jwk.ecPrv.crv,
    x = b64UrlDecode(jwk.ecPrv.d))

proc toEcPrivateKey(pem: string): EcPrivateKeyObj =
  var skCtx: SkeyDecoderContext
  skeyDecoderInit(skCtx)
  decodePem(pem, "EC PRIVATE KEY", addr skCtx,
    cast[BearPemDecoderCallback](skeyDecoderPush))

  let k = skCtx.key.ec
  return initEcPrivateKeyObj(k.curve.fromBearSslCurveConst, $cast[cstring](k.x))

proc getEcHashCfg(alg: JwtAlgorithm): HashCfg =
  let hashAlg = case alg:
    of ES256: SHA256
    of ES384: SHA384
    of ES512: SHA512
    else: raise newException(ValueError,
      "Unsupported ECDSA signature algorithm '" & $alg & "'")

  result = hashConfig(hashAlg)

proc bearEcSign(
    message: string,
    alg: JwtAlgorithm,
    key: EcPrivateKeyObj
  ): string =

  let hashCfg = getEcHashCfg(alg)
  let hashed = hash(message, hashcfg.alg)

  let ecSignImpl = ecdsaSignRawGetDefault()

  # The signature fucntion will return the length of the signature in bytes,
  # but we need to pre-allocate the space for it to write into. Per BearSSL
  # documentation, the maximum signature size will by 132 bytes (see
  # https://bearssl.org/apidoc/bearssl__ec_8h.html#ab7154b0c899ceb3af062c69016258667)
  result = newString(132)

  let sigLen = ecSignImpl(
    addr ecAllM15,
    hashCfg.vtable,
    cast[ptr char](unsafeAddr hashed[0]),
    unsafeAddr key.bearKey,
    cast[ptr char](addr result[0]))

  if sigLen == 0: raise newException(Exception, "EC signature failed")
  result.setLen(sigLen)

proc bearEcVerify(
    message, signature: string,
    alg: JwtAlgorithm,
    key: EcPublicKeyObj
  ): bool =

  let hashCfg = getEcHashCfg(alg)
  let hashed = hash(message, hashCfg.alg)

  let ecVerifyImpl = ecdsaVrfyRawGetDefault()
  let resultCode = ecVerifyImpl(
    addr ecAllM15,
    hashed.cstring,
    uint hashed.len,
    unsafeAddr key.bearKey,
    signature.cstring,
    uint signature.len)

  result = resultCode == 1

proc ecSign*(message: string, alg: JwtAlgorithm, key: JWK): string =
  ## Sign a message using the ECDSA algorithm.
  ##
  ## *key* is expected to be a `JWK <,./../jwt_full/jwk.html#JWK>`_

  if key.keyKind != JwkKeyType.EcPrivate:
    raise newException(ValueError,
      $alg & " requires an ECDSA private key (\"type\"=\"EC\"), not a " &
      "\"typ\"=\"" & $(key.keyKind) & "\" key.")

  return bearEcSign(message, alg, toEcPrivateKey(key))

proc ecSign*(message: string, alg: JwtAlgorithm, pemKey: string): string =
  ## Sign a message using the ECDSA algorithm.
  ##
  ## *pemKey* is expected to be a PEM
  return bearEcSign(message, alg, toEcPrivateKey(pemKey))

proc ecVerify*(message, signature: string, alg: JwtAlgorithm, key: JWK): bool =
  ## Verify the signature for a message using ECDSA.
  ##
  ## *key* is expected to be a `JWK <../../jwt_full/jwk.html#JWK>`_

  if key.keyKind != JwkKeyType.EcPrivate and
     key.keyKind != JwkKeyType.EcPublic:
    raise newException(ValueError,
      $alg & " requires an ECDSA private key (\"type\"=\"EC\"), not a " &
      "\"typ\"=\"" & $(key.keyKind) & "\" key.")

  return bearEcVerify(message, signature, alg, toEcPublicKey(key))

proc ecVerify*(message, signature: string, alg: JwtAlgorithm, pemKey: string): bool =
  ## Verify the signature for a message using ECDSA.
  ##
  ## *pemKey* is expected to be a PEM.
  return bearEcVerify(message, signature, alg, toEcPublicKey(pemKey))
