import std/options, std/tables
import bearssl, bearssl_pkey_decoder

import ../../encoding
import ../../jwa
import ../../jwk

import ./hash
import ./pem

type
  RsaPublicKeyObj = object
    n*, e*: string
    bearKey: RsaPublicKey

  RsaPrivateKeyObj = object
    p*, q*, dp*, dq*, iq*: string
    bearKey: RsaPrivateKey

func initRsaPublicKeyObj(n, e: string): RsaPublicKeyObj =
  result = RsaPublicKeyObj(n: n, e: e)
  result.bearKey.n = cast[ptr byte](result.n.cstring)
  result.bearKey.nlen = uint result.n.len
  result.bearKey.e = cast[ptr byte](result.e.cstring)
  result.bearKey.elen = uint result.e.len

func initRsaPrivateKeyObj(nBitLen: int, p, q, dp, dq, iq: string): RsaPrivateKeyObj =
  result = RsaPrivateKeyObj(p: p, q: q, dp: dp, dq: dq, iq: iq)
  result.bearKey.nBitLen = cast[uint32](nBitLen)
  result.bearKey.p = cast[ptr byte](result.p.cstring)
  result.bearKey.plen = uint result.p.len
  result.bearKey.q = cast[ptr byte](result.q.cstring)
  result.bearKey.qlen = uint result.q.len
  result.bearKey.dp = cast[ptr byte](result.dp.cstring)
  result.bearKey.dplen = uint result.dp.len
  result.bearKey.dq = cast[ptr byte](result.dq.cstring)
  result.bearKey.dqlen = uint result.dq.len
  result.bearKey.iq = cast[ptr byte](result.iq.cstring)
  result.bearKey.iqlen = uint result.iq.len

proc toRsaPublicKey(jwk: JWK): RsaPublicKeyObj =
  ## Convert an RSA public key in JWK format to the wrapper for BearSSL's
  ## RsaPublicKey struct.

  if jwk.keyKind != JwkKeyType.RsaPublic and
     jwk.keyKind != JwkKeyType.RsaPrivate:
    raise newException(ValueError,
      "Can not extract RSA public key from the given JWT: keyKind is " &
      $jwk.keyKind)

  return initRsaPublicKeyObj(
    n = b64UrlDecode(jwk.rsaPub.n),
    e = b64UrlDecode(jwk.rsaPub.e))


proc toRsaPublicKey(pem: string): RsaPublicKeyObj =
  var pkCtx: PkeyDecoderContext
  pkeyDecoderInit(addr pkCtx)
  decodePem(pem, "PUBLIC KEY", addr pkCtx,
    cast[BearPemDecoderCallback](pkeyDecoderPush))

  let k = pkCtx.key.rsa
  return initRsaPublicKeyObj($cast[cstring](k.n), $cast[cstring](k.e))


proc toRsaPrivateKey(jwk: JWK): RsaPrivateKeyObj =
  ## Convert an RSA private key in JWK format to the wrapper for BearSSL's
  ## RsaPrivateKey struct.
  ##
  ## Note: there are two ways to represent an rsa private key[1][rsa-priv-rep]:
  ## 1. The pair modulus and private exponent pair (*n*, *d*)
  ## 2. The quintuple with the prime factors, reduced CRT exponents, and first
  ##    CRT coefficient (*p*, *q*, *dP*, *dQ*, *qInv*)
  ##
  ## The JWA standard requires that a JWK representing an RSA private key
  ## provide rep. 1 (*n*, *d*) and optionally allows rep. 2 to allows be
  ## included.  BearSSL strictly uses rep. 2 in it's key structures.
  ## **Currently, this library only supports keys that provide the parameters
  ## for rep. 2.** Ideally this library would also support keys only containing
  ## rep. 1, including the logic to derive rep. 2 from use in BearSSL. This is
  ## a TODO.
  ##
  ## [NIST Special Publication 800-56B][NIST.SP.800-56Br2] provides a choice of
  ## two algorithms to derive the prime factors *p* and *q* from the private
  ## exponent *d*. With the prime factors known, the calculation *dP*, *dQ*,
  ## and *qInv* are straighforward based on [their definitions in RFC8
  ## 8017][rsa-apdx-a]
  ##
  ## [rsa-priv-rep]: https://datatracker.ietf.org/doc/html/rfc8017#section-3.2
  ## [rsa-apdx-a]: https://datatracker.ietf.org/doc/html/rfc8017#appendix-A.1.2
  ## [NIST.SP.800-56Br2]: https://nvlpubs.nist.gov/nistpubs/SpecialPublications/NIST.SP.800-56Br2.pdf

  if jwk.keyKind != JwkKeyType.RsaPrivate:
    raise newException(ValueError,
      "Can not extract RSA private key from the given JWT: keyKind is " &
      $jwk.keyKind)

  let sk = jwk.rsaPrv
  if sk.p.isNone or sk.q.isNone or sk.dp.isNone or sk.dq.isNone or sk.qi.isNone:
    raise newException(ValueError,
      "RSA private key must have values for: n, e, d, p, q, dp, dq, and qi")

  let n = b64UrlDecode(sk.n)
  return initRsaPrivateKeyObj(
    nBitLen = n.len * 8,
    p = b64UrlDecode(sk.p.get),
    q = b64UrlDecode(sk.q.get),
    dp = b64UrlDecode(sk.dp.get),
    dq = b64UrlDecode(sk.dq.get),
    iq = b64UrlDecode(sk.qi.get))

proc toRsaPrivateKey(pem: string): RsaPrivateKeyObj =
  var skCtx: SkeyDecoderContext
  skeyDecoderInit(skCtx)
  decodePem(pem, "RSA PRIVATE KEY", addr skCtx,
    cast[BearPemDecoderCallback](skeyDecoderPush))

  let k = skCtx.key.rsa

  return initRsaPrivateKeyObj(
    cast[int](k.nBitLen),
    $cast[cstring](k.p),
    $cast[cstring](k.q),
    $cast[cstring](k.dp),
    $cast[cstring](k.dq),
    $cast[cstring](k.iq))

proc getRsaHashCfg(alg: JwtAlgorithm): HashCfg =
  let hashAlg = case alg:
    of RS256: SHA256
    of RS384: SHA384
    of RS512: SHA512
    else: raise newException(ValueError,
      "Unsupported RSA signature algorithm '" & $alg & "'")

  result = hashConfig(hashAlg)

proc bearRsaSign(
    message: string,
    alg: JwtAlgorithm,
    key: RsaPrivateKeyObj
  ): string =

  let hashCfg = getRsaHashCfg(alg)
  let hashed = hash(message, hashCfg.alg)

  let rsaSignImpl = rsaPkcs1SignGetDefault()
  result = newString((key.bearKey.nBitLen + 7) div 8)

  let errCode = rsaSignImpl(
    cast[ptr byte](hashCfg.oid),
    cast[ptr byte](unsafeAddr hashed[0]),
    uint hashed.len,
    unsafeAddr key.bearKey,
    cast[ptr byte](addr result[0]))

  if errCode != 1: raise newException(Exception, "RSA signature failed")

proc bearRsaVerify(
    message, signature: string,
    alg: JwtAlgorithm,
    key: RsaPublicKeyObj
  ): bool =

  let hashCfg = getRsaHashCfg(alg)
  let hashed = hash(message, hashCfg.alg)

  let rsaVerifyImpl = rsaPkcs1VrfyGetDefault()
  var recoveredHash = newString(hashCfg.size)

  let errCode = rsaVerifyImpl(
    cast[ptr byte](unsafeAddr signature[0]),
    uint signature.len,
    cast[ptr byte](hashCfg.oid),
    uint hashed.len,
    unsafeAddr key.bearKey,
    cast[ptr byte](addr recoveredHash[0]))

  if errCode != 1: return false
  return hashed == recoveredHash

proc rsaSign*(message: string, alg: JwtAlgorithm, key: JWK): string =
  ## Sign a message using the RSA PKCS#1 v1.5 algorithm.
  ##
  ## *key* is expected to be a `JWK <../../jwt_full/jwk.html#JWK>`_

  if key.keyKind != JwkKeyType.RsaPrivate:
    raise newException(ValueError,
      $alg & " requires an RSA Private key (\"type\"=\"RSA\"), not a " &
      "\"typ\"=\"" & $(key.keyKind) & "\" key.")

  return bearRsaSign(message, alg, toRsaPrivateKey(key))

proc rsaSign*(message: string, alg: JwtAlgorithm, pemKey: string): string =
  ## Sign a message using hte RSA PKCS#1 v1.5 algorithm.
  ##
  ## *pemKey* is expected to be an RSA private key in PEM format.
  return bearRsaSign(message, alg, toRsaPrivateKey(pemKey))

proc rsaVerify*(message, signature: string; alg: JwtAlgorithm, key: JWK): bool =
  ## Verify the signature for a message using PKCS#1 v1.5 algorithm.
  ##
  ## *key* is expected to be a `JWK <../../jwt_full/jwk.html#JWK>`_

  if key.keyKind != JwkKeyType.RsaPublic:
    raise newException(ValueError,
      $alg & " requires an RSA Public key (\"type\"=\"RSA\"), not a " &
      "\"typ\"=\"" & $(key.keyKind) & "\" key.")

  return bearRsaVerify(message, signature, alg, toRsaPublicKey(key))

proc rsaVerify*(message, signature: string; alg: JwtAlgorithm, pemKey: string): bool =
  ## Verify the signature for a message using PKCS#1 v1.5 algorithm.
  ##
  ## *key* is expected to be an RSA public key in PEM format.
  return bearRsaVerify(message, signature, alg, toRsaPublicKey(pemKey))
